#include "config.h"
#include "eval.h"
#include "position.h"
#include "search.h"

int WeightMaterial    = 100;
int WeightPSQ         = 100;
int WeightMobility    = 100;
int WeightKingSafety  = 100;
int WeightPawns       = 100;
int WeightPassedPawns = 100;

const EVAL LAZY_MARGIN                 = VAL_BISHOP;

const EVAL QUEEN_K_TROPISM             = 1;
const EVAL QUEEN_ON_7TH                = 8;

const EVAL ROOK_ON_7TH                 = 16;
const EVAL ROOK_ON_OPEN                = 8;
const EVAL ROOK_ON_SEMIOPEN            = 4;

const EVAL PAWN_MOBILITY               = 1;
const EVAL KNIGHT_MOBILITY             = 3;
const EVAL BISHOP_MOBILITY             = 4;
const EVAL ROOK_MOBILITY               = 2;

const EVAL BISHOP_TRAPPED              = 64;

const EVAL PAWN_SHIELD_OWN_3RD         = 8;
const EVAL PAWN_SHIELD_OWN_4TH         = 16;
const EVAL PAWN_SHIELD_OWN_NOPAWN      = 24;

const EVAL PAWN_SHIELD_OPP_NOPAWN      = 16;
const EVAL PAWN_SHIELD_OPP_3RD         = 8;
const EVAL PAWN_SHIELD_OPP_4TH         = 4;

const EVAL PAWN_PASSED_BASE            = 16;
const EVAL PAWN_PASSED_UNCATCHABLE     = 64;
const EVAL PAWN_PASSED_FAR_PAIR        = 128;

const EVAL PAWN_DOUBLED                = 16;
const EVAL PAWN_ISOLATED               = 8;
const EVAL PAWN_BACKWARDS              = 4;

const EVAL TEMPO                       = 12;

static const FLD FLIP[64] =
{
	A1, B1, C1, D1, E1, F1, G1, H1,
	A2, B2, C2, D2, E2, F2, G2, H2,
	A3, B3, C3, D3, E3, F3, G3, H3,
	A4, B4, C4, D4, E4, F4, G4, H4,
	A5, B5, C5, D5, E5, F5, G5, H5,
	A6, B6, C6, D6, E6, F6, G6, H6,
	A7, B7, C7, D7, E7, F7, G7, H7,
	A8, B8, C8, D8, E8, F8, G8, H8
};

EVAL PSQ_KING[64] =
{
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-30, -30, -30, -30, -30, -30, -30, -30,
	-20, -20, -20, -20, -20, -20, -20, -20,
	-10,   0,   0, -20,  -5, -20,   0, -10,
};

EVAL PSQ_KING_ENDGAME[64] =
{
	-20, -10, -10, -10, -10, -10, -20, -20,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   0,  10,  10,  10,  10,   0, -10,
	-10,   0,  10,  20,  20,  10,   0, -10,
	-10,   0,  10,  20,  20,  10,   0, -10,
	-10,   0,  10,  10,  10,  10,   0, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-20, -10, -10, -10, -10, -10, -10, -20
};

EVAL PSQ_BISHOP[64] =
{
	 -9,  -6,  -3,   0,   0,  -3,  -6,  -9,
	 -6,  -3,   0,   3,   3,   0,  -3,  -6,
	 -3,   0,   3,   6,   6,   3,   0,  -3,
	  0,   3,   6,   9,   9,   6,   3,   0,
	  0,   3,   6,   9,   9,   6,   3,   0,
	 -3,   0,   3,   6,   6,   3,   0,  -3,
	 -6,  -3,   0,   3,   3,   0,  -3,  -6,
	 -9,  -6,  -3,   0,   0,  -3,  -6,  -9,
};

EVAL PSQ_KNIGHT[64] =
{
	-12,  -8,  -4,   0,   0,  -4,  -8, -12,
	 -8,  -4,   0,   4,   4,   0,  -4,  -8,
	 -4,   0,   4,   8,   8,   4,   0,  -4,
	  0,   4,   8,  12,  12,   8,   4,   0,
	  0,   4,   8,  12,  12,   8,   4,   0,
	 -4,   0,   4,   8,   8,   4,   0,  -4,
	 -8,  -4,   0,   4,   4,   0,  -4,  -8,
	-12,  -8,  -4,   0,   0,  -4,  -8, -12
};

EVAL PSQ_PAWN[64] =
{
	  0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   2,   2,   2,   2,   0,   0,
	  0,   2,   4,   6,   6,   4,   2,   0,
	  0,   2,   6,   8,   8,   6,   2,   0,
	  0,   2,   6,   6,   6,   6,   2,   0,
	  0,   2,   2,   2,   2,   2,   2,   0,
	  0,   0,   0,   0,   0,   0,   0,   0
};

EVAL EvalPawnShieldW(const PawnHashEntry* pentry, FLD  K);
EVAL EvalPawnShieldB(const PawnHashEntry* pentry, FLD  K);

EVAL Evaluate(const Position& pos, EVAL alpha, EVAL beta)
{
	U64 x = 0;
	FLD f = NF;
	U64 occupied = pos.BitsAll();
	U64 free = ~occupied;

	if (pos.IsDraw())
		return DRAW_SCORE;

	EVAL mat_score = pos.Material(WHITE) - pos.Material(BLACK);
	if (pos.Count(BW) == 2)
		mat_score += BISHOP_PAIR;
	if (pos.Count(BB) == 2)
		mat_score -= BISHOP_PAIR;

	EVAL ksf_score = 0;
	EVAL pas_score = 0;
	EVAL psq_score = 0;
	EVAL mob_score = 0;
	EVAL pwn_score = 0;

	//
	//   LAZY EVAL
	//

	EVAL e = mat_score;
	if (pos.Side() == BLACK)
		e = - e;

	if (e > beta + LAZY_MARGIN)
		return beta;

	if (e < alpha - LAZY_MARGIN)
		return alpha;

	//
	//   MATERIAL COEFFICIENTS
	//

	EVAL king_safety[2] = {0, 0};

	U32 ph = pos.PawnHash();
	long index = ph % g_pawn_hash_size;
	PawnHashEntry* pentry = & (g_pawn_hash[index]);

	if (pentry->m_hash != ph)
		pentry->Read(pos);

	//
	//   KINGS
	//

	if (pos.MatIndex(BLACK) >= 10)
	{
		king_safety[WHITE] += PSQ_KING[pos.King(WHITE)];
		king_safety[WHITE] += EvalPawnShieldW(pentry, pos.King(WHITE));
	}
	else
		psq_score += PSQ_KING_ENDGAME[pos.King(WHITE)];

	if (pos.MatIndex(WHITE) >= 10)
	{
		king_safety[BLACK] += PSQ_KING[FLIP[pos.King(BLACK)]];
		king_safety[BLACK] += EvalPawnShieldB(pentry, pos.King(BLACK));
	}
	else
		psq_score -= PSQ_KING_ENDGAME[FLIP[pos.King(BLACK)]];

	//
	//   PAWNS
	//

	pwn_score = pentry->m_score;

	x = Up(pos.Bits(PW)) & free;
	x |= pentry->m_attackedBy[WHITE] & pos.BitsAll(BLACK);
	mob_score += PAWN_MOBILITY * CountBits(x);

	x = Down(pos.Bits(PB)) & free;
	x |= pentry->m_attackedBy[BLACK] & pos.BitsAll(WHITE);
	mob_score -= PAWN_MOBILITY * CountBits(x);

	// passed pawns

	x = pentry->m_passed & pos.Bits(PW);
	while (x)
	{
		f = PopLSB(x);
		pas_score += PAWN_PASSED_BASE * (7 - Row(f));

		if (pos.MatIndex(BLACK) == 0)
		{
			if (pos.Side() == BLACK)
				f += 8;

			if ((BB_PAWN_SQUARE[f][WHITE] & pos.Bits(KB)) == 0)
				pas_score += PAWN_PASSED_UNCATCHABLE * (7 - Row(f));
		}
	}

	x = (pentry->m_passed & pos.Bits(PW)) & (BB_A7H7 | BB_A6H6);
	x = (x | Up(x)) & BB_A7H7;
	x = x & Right(x);
	pas_score += PAWN_PASSED_FAR_PAIR * CountBits(x);

	x = pentry->m_passed & pos.Bits(PB);
	while (x)
	{
		f = PopLSB(x);
		pas_score -= PAWN_PASSED_BASE * Row(f);

		if (pos.MatIndex(WHITE) == 0)
		{
			if (pos.Side() == WHITE)
				f -= 8;

			if ((BB_PAWN_SQUARE[f][BLACK] & pos.Bits(KW)) == 0)
				pas_score -= PAWN_PASSED_UNCATCHABLE * Row(f);
		}
	}

	x = (pentry->m_passed & pos.Bits(PB)) & (BB_A2H2 | BB_A3H3);
	x = (x | Down(x)) & BB_A2H2;
	x = x & Right(x);
	pas_score -= PAWN_PASSED_FAR_PAIR * CountBits(x);

	//
	//   KNIGHTS
	//

	x = pos.Bits(NW);
	while (x)
	{
		f = PopLSB(x);
		psq_score += PSQ_KNIGHT[f];
		U64 y = BB_KNIGHT_ATTACKS[f];
		mob_score += KNIGHT_MOBILITY * CountBits(y);
	}

	x = pos.Bits(NB);
	while (x)
	{
		f = PopLSB(x);
		psq_score -= PSQ_KNIGHT[FLIP[f]];
		U64 y = BB_KNIGHT_ATTACKS[f];
		mob_score -= KNIGHT_MOBILITY * CountBits(y);
	}

	//
	//   BISHOPS
	//

	x = pos.Bits(BW);
	while (x)
	{
		f = PopLSB(x);
		psq_score += PSQ_BISHOP[f];
		U64 y = BishopAttacks(f, occupied);
		mob_score += BISHOP_MOBILITY * CountBits(y);

		if (f == A7 && pos[B6] == PB)
			mob_score -= BISHOP_TRAPPED;
		if (f == H7 && pos[G6] == PB)
			mob_score -= BISHOP_TRAPPED;
	}

	x = pos.Bits(BB);
	while (x)
	{
		f = PopLSB(x);
		psq_score -= PSQ_BISHOP[FLIP[f]];
		U64 y = BishopAttacks(f, occupied);
		mob_score -= BISHOP_MOBILITY * CountBits(y);

		if (f == A2 && pos[B3] == PW)
			psq_score += BISHOP_TRAPPED;
		if (f == H2 && pos[G3] == PW)
			psq_score += BISHOP_TRAPPED;
	}

	//
	//   ROOKS
	//

	x = pos.Bits(RW);
	while (x)
	{
		f = PopLSB(x);
		int file = Col(f) + 1;

		if (Row(f) == 1 && Row(pos.King(BLACK)) <= 1)
			psq_score += ROOK_ON_7TH;
		
		if (pentry->m_ranks[WHITE][file] == 0)
		{
			if (pentry->m_ranks[BLACK][file] == 7)
				psq_score += ROOK_ON_OPEN;
			else
				psq_score += ROOK_ON_SEMIOPEN;
		}

		U64 y = RookAttacks(f, occupied);
		mob_score += ROOK_MOBILITY * CountBits(y);
	}

	x = pos.Bits(RB);
	while (x)
	{
		f = PopLSB(x);
		int file = Col(f) + 1;

		if (Row(f) == 6 && Row(pos.King(WHITE)) >= 6)
			psq_score -= ROOK_ON_7TH;

		if (pentry->m_ranks[BLACK][file] == 7)
		{
			if (pentry->m_ranks[WHITE][file] == 0)
				psq_score -= ROOK_ON_OPEN;
			else
				psq_score -= ROOK_ON_SEMIOPEN;
		}

		U64 y = RookAttacks(f, occupied);
		mob_score -= ROOK_MOBILITY * CountBits(y);
	}

	//
	//   QUEENS
	//

	x = pos.Bits(QW);
	while (x)
	{
		f = PopLSB(x);
		if (Row(f) == 1 && Row(pos.King(BLACK)) <= 1)
			psq_score += QUEEN_ON_7TH;

		int dist = g_dist[f][pos.King(BLACK)];
		king_safety[BLACK] -= QUEEN_K_TROPISM * (7 - dist);
	}

	x = pos.Bits(QB);
	while (x)
	{
		f = PopLSB(x);
		if (Row(f) == 6 && Row(pos.King(WHITE)) >= 6)
			psq_score -= QUEEN_ON_7TH;

		int dist = g_dist[f][pos.King(WHITE)];
		king_safety[WHITE] -= QUEEN_K_TROPISM * (7 - dist);
	}

	//
	//   SCALING
	//

	ksf_score = (king_safety[WHITE] * pos.MatIndex(BLACK) - king_safety[BLACK] * pos.MatIndex(WHITE)) / 32;

	mat_score = mat_score * WeightMaterial;
	psq_score = psq_score * WeightPSQ;
	mob_score = mob_score * WeightMobility;
	ksf_score = ksf_score * WeightKingSafety;
	pwn_score = pwn_score * WeightPawns;
	pas_score = pas_score * WeightPassedPawns;

	EVAL score = (mat_score + ksf_score + pas_score + psq_score + mob_score + pwn_score) / 100;

	//
	//   NO-WIN MATERIAL
	//

	if (pos.Count(PW) == 0 && pos.MatIndex(WHITE) < 5)
	{
		if (score > DRAW_SCORE)
			score = DRAW_SCORE;
	}

	if (pos.Count(PB) == 0 && pos.MatIndex(BLACK) < 5)
	{
		if (score < DRAW_SCORE)
			score = DRAW_SCORE;
	}

	if (pos.Side() == BLACK)
		score = -score;

	return score + TEMPO;
}
////////////////////////////////////////////////////////////////////////////////

void PawnHashEntry::Read(const Position& pos)
{
	m_hash = pos.PawnHash();
	m_passed = 0;
	m_score = 0;

	U64 x = 0, y = 0;
	FLD f = NF;

	for (int file = 0; file < 10; file++)
	{
		m_ranks[WHITE][file] = 0;
		m_ranks[BLACK][file] = 7;
	}

	x = pos.Bits(PW);
	y = Up(x);
	m_attackedBy[WHITE] = Left(y) | Right(y);

	while (x)
	{
		f = PopLSB(x);
		m_score += PSQ_PAWN[f];

		int file = Col(f) + 1;
		int rank = Row(f);

		if (rank > m_ranks[WHITE][file])
			m_ranks[WHITE][file] = rank;

		y = BB_DIR[f][DIR_U];
		y |= Left(y);
		y |= Right(y);

		if ((y & pos.Bits(PB)) == 0)
			m_passed |= BB_SINGLE[f];
	}

	x = pos.Bits(PB);
	y = Down(x);
	m_attackedBy[BLACK] = Left(y) | Right(y);

	while (x)
	{
		f = PopLSB(x);
		m_score -= PSQ_PAWN[FLIP[f]];

		int file = Col(f) + 1;
		int rank = Row(f);

		if (rank < m_ranks[BLACK][file])
			m_ranks[BLACK][file] = rank;

		y = BB_DIR[f][DIR_D];
		y |= Left(y);
		y |= Right(y);

		if ((y & pos.Bits(PW)) == 0)
			m_passed |= BB_SINGLE[f];
	}

	// Second pass

	x = pos.Bits(PW);
	while (x)
	{
		f = PopLSB(x);
		int file = Col(f) + 1;
		int rank = Row(f);

		if (m_ranks[WHITE][file] != rank)
			m_score -= PAWN_DOUBLED;

		if (m_ranks[WHITE][file - 1] == 0 && m_ranks[WHITE][file + 1] == 0)
			m_score -= PAWN_ISOLATED;
		else if (m_ranks[WHITE][file - 1] < rank && m_ranks[WHITE][file + 1] < rank)
			m_score -= PAWN_BACKWARDS;
	}

	x = pos.Bits(PB);
	while (x)
	{
		f = PopLSB(x);
		int file = Col(f) + 1;
		int rank = Row(f);

		if (m_ranks[BLACK][file] != rank)
			m_score += PAWN_DOUBLED;

		if (m_ranks[BLACK][file - 1] == 7 && m_ranks[BLACK][file + 1] == 7)
			m_score += PAWN_ISOLATED;
		else if (m_ranks[BLACK][file - 1] > rank && m_ranks[BLACK][file + 1] > rank)
			m_score += PAWN_BACKWARDS;
	}
}
////////////////////////////////////////////////////////////////////////////////

EVAL EvalWKP(const PawnHashEntry* pentry, int file)
{
	EVAL ret = 0;

	int rank = pentry->m_ranks[WHITE][file];
	if (rank == 6);        // pawn not moved - OK
	else if (rank == 5)    // pawn on 6th
		ret -= PAWN_SHIELD_OWN_3RD;
	else if (rank != 0)    // pawn on 5th or more far
		ret -= PAWN_SHIELD_OWN_4TH;
	else
		ret -= PAWN_SHIELD_OWN_NOPAWN; // no pawn at all

	rank = pentry->m_ranks[BLACK][file];
	if (rank == 7)         // no enemy's pawn at all
		ret -= PAWN_SHIELD_OPP_NOPAWN;
	else if (rank == 5)
		ret -= PAWN_SHIELD_OPP_3RD;
	else if (rank == 4)
		ret -= PAWN_SHIELD_OPP_4TH;

	return ret;
}
////////////////////////////////////////////////////////////////////////////////

EVAL EvalBKP(const PawnHashEntry* pentry, int file)
{
	int rank = pentry->m_ranks[BLACK][file];
	EVAL ret = 0;

	if (rank == 1);        // pawn not moved - OK
	else if (rank == 2)    // pawn on 6th
		ret -= PAWN_SHIELD_OWN_3RD;
	else if (rank != 7)    // pawn on 5th or more far
		ret -= PAWN_SHIELD_OWN_4TH;
	else
		ret -= PAWN_SHIELD_OWN_NOPAWN; // no pawn at all

	rank = pentry->m_ranks[WHITE][file];

	if (rank == 0)         // no enemy's pawn at all
		ret -= PAWN_SHIELD_OPP_NOPAWN;
	else if (rank == 2)
		ret -= PAWN_SHIELD_OPP_3RD;
	else if (rank == 3)
		ret -= PAWN_SHIELD_OPP_4TH;

	return ret;
}
////////////////////////////////////////////////////////////////////////////////

EVAL EvalPawnShieldW(const PawnHashEntry* pentry, FLD  K)
{
	int file = Col(K) + 1;
	EVAL ret = 0;

	for (int j = file - 1; j <= file + 1; ++j)
		ret += EvalWKP(pentry, j);

	return ret;
}
////////////////////////////////////////////////////////////////////////////////

EVAL EvalPawnShieldB(const PawnHashEntry* pentry, FLD  K)
{
	int file = Col(K) + 1;
	EVAL ret = 0;

	for (int j = file - 1; j <= file + 1; ++j)
		ret += EvalBKP(pentry, j);

	return ret;
}
////////////////////////////////////////////////////////////////////////////////

void InitEval()
{
	PARAM(WeightMaterial)
	PARAM(WeightPSQ)
	PARAM(WeightMobility)
	PARAM(WeightKingSafety)
	PARAM(WeightPawns)
	PARAM(WeightPassedPawns)
}
////////////////////////////////////////////////////////////////////////////////

